@@ -9,7 +9,18 @@
 #include "pm.h"
 #include "pnm.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+#if 0
+} /* to fake out automatic code indenters */
+#endif
+
 typedef unsigned long sample;
+    /* Regardless of the capacity of "unsigned long", a sample is always
+       less than 1 << 16.  This is essential for some code to avoid
+       arithmetic overflows.
+    */
 
 struct pam {
 /* This structure describes an open PAM image file.  It consists
@@ -34,17 +45,23 @@ struct pam {
            The information starts in the first byte and is contiguous.  
            This cannot be greater than 'size'
            */
-    FILE *file;
+    FILE * file;
     int format;
         /* The format code of the raw image.  This is PAM_FORMAT
            unless the PAM image is really a view of a PBM, PGM, or PPM
            image.  Then it's PBM_FORMAT, RPBM_FORMAT, etc.
            */
     unsigned int plainformat;
-        /* Logical:  the format above is a plain (text) format as opposed
-           to a raw (binary) format.  This is entirely redundant with the
-           'format' member and exists as a separate member only for 
-           computational speed.
+        /* Logical: On output, use the plain version of the format type
+           indicated by 'format'.  Otherwise, use the raw version.
+           (i.e., on output, the plainness information in 'format' is
+           irrelevant).  Input functions set this to FALSE, for the
+           convenience of programs that copy an input pam structure for
+           use with output.
+
+           Before Netpbm 10.32, this was rather different.  It simply
+           described for convenience the plainness of the format indicated
+           by 'format'.
         */
     int height;  /* Height of image in rows */
     int width;   
@@ -75,14 +92,37 @@ struct pam {
 
            0 means the allocation depth is the same as the image depth.
         */
+    const char ** comment_p;
+        /* Pointer to a pointer to a NUL-terminated ASCII string of
+           comments.  When reading an image, this contains the
+           comments from the image's PAM header; when writing, the
+           image gets these as comments, right after the magic number
+           line.  The individual comments are delimited by newlines
+           and are in the same order as in the PAM header.
+
+           On output, NULL means no comments.
+
+           On input, libnetpbm mallocs storage for the comments and placed
+           the pointer at *comment_p.  Caller must free it.  NULL means
+           libnetpbm does not return comments and does not allocate any
+           storage.
+        */
 };
 
+#define PAM_HAVE_ALLOCATION_DEPTH 1
+#define PAM_HAVE_COMMENT_P 1
+
 /* PAM_STRUCT_SIZE(x) tells you how big a struct pam is up through the 
    member named x.  This is useful in conjunction with the 'len' value
    to determine which fields are present in the structure.
 */
+
+/* Some compilers are really vigilant and recognize it as an error
+   to cast a 64 bit address to a 32 bit type.  Hence the roundabout
+   casting in PAM_MEMBER_OFFSET.
+*/
 #define PAM_MEMBER_OFFSET(mbrname) \
-  ((unsigned int)(char*)&((struct pam *)0)->mbrname)
+  ((size_t)(unsigned long)(char*)&((struct pam *)0)->mbrname)
 #define PAM_MEMBER_SIZE(mbrname) \
   sizeof(((struct pam *)0)->mbrname)
 #define PAM_STRUCT_SIZE(mbrname) \
@@ -126,7 +166,7 @@ typedef sample *tuple;
        of the image (see pam.depth above).
     */
 
-#define PAM_OVERALL_MAXVAL ULONG_MAX
+#define PAM_OVERALL_MAXVAL 65535
 
 /* Note: xv uses the same "P7" signature for its thumbnail images (it
    started using it years before PAM and unbeknownst to the designer
@@ -148,15 +188,39 @@ struct pamtuples {
 };
 
 
+typedef float * pnm_transformMap;
+    /* This is an array of transform maps.  transform[N] is the 
+       array that is the map for Plane N.
+   
+       Transform maps define a transformation between PAM sample value
+       to normalized libnetpbm "samplen" value, i.e. what you get back
+       from pnm_readpamrown() or pass to pnm_writepamrown().
+       Typically, it's a gamma transfer function generated by
+       pnm_creategammatransform() or pnm_createungammatransform().
+
+       NULL for any transform means just plain normalization -- divide
+       the PAM sample value by the maxval to get the samplen, multiply
+       samplen by the maxval and round to get PAM sample value.
+
+       NULL for map table, or 'transform' member not present (pam
+       structure is too small to contain it) means ALL transforms
+       are plain normalization.
+
+       Each transform map is an array indexed by a PAM sample
+       value, containing 'float' values.  So it must have 'maxval'
+       entries.  The sample -> samplen tranformation is just the
+       obvious table lookup.  The samplen -> sample transformation is
+       more complicated -- if the samplen value is between map[N]
+       and map[N+1], then the sample value is N.  And only transforms
+       where map[N+1] > map[N] are allowed.  
+    */
+
 /* Declarations of library functions. */
 
 /* We don't have a specific PAM function for init and nextimage, because
    one can simply use pnm_init() and pnm_nextimage() from pnm.h.
 */
 
-char
-stripeq(const char * const comparand, const char * const comparator);
-
 unsigned int
 pnm_bytespersample(sample const maxval);
 
@@ -210,7 +274,7 @@ pnm_makearrayrgb(const struct pam * const pamP,
 
 void
 pnm_getopacity(const struct pam * const pamP,
-               bool *             const haveOpacityP,
+               int *              const haveOpacityP,
                unsigned int *     const opacityPlaneP);
 
 void
@@ -252,13 +316,17 @@ void
 pnm_freerowimage(unsigned char * const rowimage);
 
 void 
-pnm_readpaminit(FILE *file, struct pam * const pamP, const int size);
+pnm_readpaminit(FILE *       const file, 
+                struct pam * const pamP, 
+                int          const size);
 
 void 
 pnm_readpamrow(const struct pam * const pamP, tuple* const tuplerow);
 
-tuple** 
-pnm_readpam(FILE *file, struct pam * const pamP, const int size);
+tuple ** 
+pnm_readpam(FILE *       const file, 
+            struct pam * const pamP, 
+            int          const size);
 
 void 
 pnm_writepaminit(struct pam * const pamP);
@@ -282,7 +350,7 @@ pnm_writepam(struct pam * const pamP, tuple ** const tuplearray);
 
 void
 pnm_checkpam(const struct pam *   const pamP, 
-             const enum pm_check_type   check_type, 
+             enum pm_check_type   const checkType,
              enum pm_check_code * const retvalP);
 
 /*----------------------------------------------------------------------------
@@ -337,9 +405,20 @@ pnm_normalizetuple(struct pam * const pamP,
 
 void
 pnm_unnormalizetuple(struct pam * const pamP,
-                    tuplen        const tuplen,
-                    tuple         const tuple);
+                     tuplen       const tuplen,
+                     tuple        const tuple);
 
+void
+pnm_normalizeRow(struct pam *             const pamP,
+                 const tuple *            const tuplerow,
+                 const pnm_transformMap * const transform,
+                 tuplen *                 const tuplenrow);
+
+void
+pnm_unnormalizeRow(struct pam *             const pamP,
+                   const tuplen *           const tuplenrow,
+                   const pnm_transformMap * const transform,
+                   tuple *                  const tuplerow);
 
 /*----------------------------------------------------------------------------
    Facilities for working with visual images in particular
@@ -362,6 +441,18 @@ void
 pnm_unapplyopacityrown(struct pam * const pamP,
                        tuplen *     const tuplenrow);
 
+pnm_transformMap *
+pnm_creategammatransform(const struct pam * const pamP);
+
+void
+pnm_freegammatransform(const pnm_transformMap * const transform,
+                       const struct pam *       const pamP);
+
+pnm_transformMap *
+pnm_createungammatransform(const struct pam * const pamP);
+
+#define pnm_freeungammatransform pnm_freegammatransform;
+
 tuple
 pnm_parsecolor(const char * const colorname,
                sample       const maxval);
@@ -381,6 +472,9 @@ pnm_YCbCr_to_rgbtuple(const struct pam * const pamP,
                       double             const Cr,
                       int *              const overflowP);
 
+#define pnm_rgbtupleisgray(tuple) \
+    ((tuple)[PAM_RED_PLANE] == (tuple)[PAM_GRN_PLANE] && \
+     (tuple)[PAM_RED_PLANE] == (tuple)[PAM_BLU_PLANE])
 
 /*----------------------------------------------------------------------------
    These are meant for passing to pm_system() as Standard Input feeder
@@ -395,4 +489,7 @@ void
 pm_accept_to_pamtuples(int    const pipeToSuckFd,
                        void * const accepterParm);
 
+#ifdef __cplusplus
+}
+#endif
 #endif
